/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2004-2010 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Application
    pisoFoam

Description
    Transient solver for incompressible flow.

    Turbulence modelling is generic, i.e. laminar, RAS or LES may be selected.

\*---------------------------------------------------------------------------*/

#include "fvCFD.H"
#include "singlePhaseTransportModel.H"
//#include "turbulenceModel.H"
#include "OFstream.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

int main(int argc, char *argv[])
{
    #include "setRootCase.H"
    #include "createTime.H"
    #include "createMesh.H"
    #include "createFields.H"
    #include "initContinuityErrs.H"

    #include "readTransportProperties.H"
    // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

    Info<< "Calculating initial conditions" << endl;
    // Initial conditions based on analytical solution at t=0
    forAll(U,cellI)
    {
        U[cellI].x()=-Foam::cos(U.mesh().C()[cellI].x())
        *Foam::sin(U.mesh().C()[cellI].y());
        U[cellI].y()=Foam::sin(U.mesh().C()[cellI].x())
        *Foam::cos(U.mesh().C()[cellI].y());
        p[cellI]=-0.25*(Foam::cos(2*U.mesh().C()[cellI].x())
        +Foam::cos(2*U.mesh().C()[cellI].y()));
    }
    // Error of the velocity component
    volVectorField errorVelocity
    (
        IOobject
        (
            "errorVelocity",
            runTime.timeName(),
            mesh,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        mesh,
        dimensionedVector("errorVelocity",dimVelocity,vector::zero)
    );

    // Analytical Solution for Velocity
    volVectorField analyticalU
    (
        IOobject
        (
            "analyticalU",
            runTime.timeName(),
            mesh,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        mesh,
        dimensionedVector("analyticalU",dimVelocity,vector::zero)
    );
    // Analytical Solution for Pressure
    volScalarField analyticalp
    (
        IOobject
        (
            "analyticalp",
            runTime.timeName(),
            mesh,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        mesh,
        dimensionedScalar("analyticalp",dimPressure,0)
    );

    DynamicList<scalar> E_rms_u;
    DynamicList<scalar> E_rms_v;
    DynamicList<scalar> E_max_u;
    DynamicList<scalar> E_max_v;

    scalar maxError;

    OFstream sPtr1(runTime.path()/"Error");

    sPtr1<<"# Time(s)"<<" "<<"dx"<<" "<<"Error"<< " "<<endl;

    Info<< "\nStarting time loop\n" << endl;

    while (runTime.loop())
    {
        Info<< "Time = " << runTime.timeName() << nl << endl;

        #include "readPISOControls.H"
        #include "CourantNo.H"
        // Analytical Solution
    forAll(analyticalU,cellI)
    {
        analyticalU[cellI].x()=-Foam::cos(U.mesh().C()[cellI].x())
        *Foam::sin(U.mesh().C()[cellI].y())*Foam::exp(-2.0*runTime.value());
        analyticalU[cellI].y()= Foam::sin(U.mesh().C()[cellI].x())
        *Foam::cos(U.mesh().C()[cellI].y())*Foam::exp(-2.0*runTime.value());
        analyticalp[cellI]=-0.25*(Foam::cos(2*U.mesh().C()[cellI].x())
        +Foam::cos(2*U.mesh().C()[cellI].y()))*Foam::exp(-4.0*runTime.value());
    }

        // Pressure-velocity PISO corrector
        {
            // Momentum predictor

            fvVectorMatrix UEqn
            (
                fvm::ddt(U)
              + fvm::div(phi, U)
//              + turbulence->divDevReff(U)
              - fvm::laplacian(nu, U)

            );

            UEqn.relax();

            if (momentumPredictor)
            {
                solve(UEqn == -fvc::grad(p));
            }

            // --- PISO loop

            for (int corr=0; corr<nCorr; corr++)
            {
                volScalarField rAU(1.0/UEqn.A());

                U = rAU*UEqn.H();
                phi = (fvc::interpolate(U) & mesh.Sf())
                    + fvc::ddtPhiCorr(rAU, U, phi);

                adjustPhi(phi, U, p);

                // Non-orthogonal pressure corrector loop
                for (int nonOrth=0; nonOrth<=nNonOrthCorr; nonOrth++)
                {
                    // Pressure corrector

                    fvScalarMatrix pEqn
                    (
                        fvm::laplacian(rAU, p) == fvc::div(phi)
                    );

                    pEqn.setReference(pRefCell, pRefValue);

                    if
                    (
                        corr == nCorr-1
                     && nonOrth == nNonOrthCorr
                    )
                    {
                        pEqn.solve(mesh.solver("pFinal"));
                    }
                    else
                    {
                        pEqn.solve();
                    }

                    if (nonOrth == nNonOrthCorr)
                    {
                        phi -= pEqn.flux();
                    }
                }

                #include "continuityErrs.H"

                U -= rAU*fvc::grad(p);
                U.correctBoundaryConditions();
            }
        }

//        turbulence->correct();

        errorVelocity=U-analyticalU;

maxError=mag(errorVelocity[0]);
forAll(errorVelocity,cellI)
{
if (mag(errorVelocity[cellI])>maxError)
{
maxError=mag(errorVelocity[cellI]);
}
}

        sPtr1<<runTime.value()<<" "<< U.mesh().C()[1].x() -
        U.mesh().C()[0].x() <<" "<<maxError<< " "<<endl;

        runTime.write();

        Info<< "ExecutionTime = " << runTime.elapsedCpuTime() << " s"
            << "  ClockTime = " << runTime.elapsedClockTime() << " s"
            << nl << endl;
    }

    Info<< "End\n" << endl;

    return 0;
}


// ************************************************************************* //
